#include "DSC-100.h"

extern int *stuck;

void run_mov(struct REGS *regs, unsigned int ip, unsigned char *mem);
void run_call(struct REGS *regs, unsigned int *ip, unsigned char *mem);
void run_ret(struct REGS *regs, unsigned int *ip, unsigned char *mem);
void run_jmp(struct REGS *regs, unsigned int *ip, unsigned char *mem);
void run_add(struct REGS *regs, unsigned int ip, unsigned char *mem);

void init_regs(struct REGS *regs)
{
	int i; 

	regs->flags = 0;
	regs->ip = 0;
	for(i = 0; i < 256; i++) regs->regs[i] = 0;
	regs->sp = 0;
}

int init_disk(unsigned char *mem, char *filename)
{
	HANDLE hFile;
	DWORD size;
	DWORD sizeRead;

	hFile=CreateFile(filename,GENERIC_READ,0,NULL,
		OPEN_EXISTING,FILE_ATTRIBUTE_NORMAL,0);
	if(hFile==INVALID_HANDLE_VALUE)return -1;
	size=GetFileSize(hFile,NULL);
	if(size==0xFFFFFFFF) {
		CloseHandle(hFile);
		return -1;
	}

	ReadFile(hFile,(void*)(mem+DISKBUF),size,&sizeRead,NULL);
	CloseHandle(hFile);

	return 0;

}

void run_bin(struct REGS *regs, unsigned char *mem)
{
	unsigned int ip = regs->ip;
	
	if(mem[ip] == NOP) {
		ip++;
	} else if(mem[ip] == MOV) {
		run_mov(regs, ip, mem);
		ip += 13;
	} else if(mem[ip] == CALL) {
		run_call(regs, &ip, mem);
	} else if(mem[ip] == RET) {
		run_ret(regs, &ip, mem);
		ip += 15;
	} else if(mem[ip] == JMP) {
		run_jmp(regs, &ip, mem);
	} else if(mem[ip] == ADD) {
		run_add(regs, ip, mem);
		ip += 13;
	}
	regs->ip = ip;
}

void run_mov(struct REGS *regs, unsigned int ip, unsigned char *mem)
{
	struct IP_MOV *mov = (struct IP_MOV *) mem + 0x1000;
	unsigned int *mov0 = (unsigned int *) mem + 0x1010, *mov1 = (unsigned int *) mem + 0x1020;

	mov->data1 = *((unsigned int *) &mem[ip + 9]);
	mov->bincode = mem[ip];
	mov->mrflug00 = mem[ip + 1];
	mov->mrflug01 = mem[ip + 2];
	mov->mrflug10 = mem[ip + 7];
	mov->mrflug11 = mem[ip + 8];
	mov->data0 = *((unsigned int *) &mem[ip + 3]);

	if(mov->mrflug00 == MOV_REG && mov->mrflug01 == MOV_REG) {
		mov0 = &regs->regs[mov->data0 - REGS_BASE];
	} else if(mov->mrflug00 == MOV_MEM && mov->mrflug01 == MOV_REG) {
		if(mov->data0 - REGS_BASE <= 257) mov0 = (unsigned int *) &mem[regs->regs[mov->data0 - REGS_BASE]];
		else mov0 = (unsigned int *) &mem[mov->data0];
	} else if(mov->mrflug00 == MOV_ETC && mov->mrflug01 == MOV_ETC) {
		mov0 = (unsigned int *) mem;
		*mov0 = mov->data0;
	} else if(mov->mrflug00 == MOV_MEM && mov->mrflug01 == MOV_ETC) {
		mov0 = (unsigned int *) &mem[mov->data0];
	}

	if(mov->mrflug10 == MOV_REG && mov->mrflug11 == MOV_REG) {
		mov1 = &regs->regs[mov->data1 - REGS_BASE];
	} else if(mov->mrflug10 == MOV_MEM && mov->mrflug11 == MOV_REG) {
		if(mov->data1 - REGS_BASE <= 257) mov1 = (unsigned int *) &mem[regs->regs[mov->data1 - REGS_BASE]];
		else mov1 = (unsigned int *) &mem[mov->data1];
	} else if(mov->mrflug10 == MOV_ETC && mov->mrflug11 == MOV_ETC) {
		mov1 = (unsigned int *) mem;
		*mov1 = mov->data1;
	} else if(mov->mrflug10 == MOV_MEM && mov->mrflug11 == MOV_ETC) {
		mov1 = (unsigned int *) &mem[mov->data1];
	}

	*mov0 = *mov1;
}

void run_call(struct REGS *regs, unsigned int *ip, unsigned char *mem)
{
	struct IP_CALL *call = (struct IP_CALL *) mem + 0x1000;
	unsigned int *addr = (unsigned int *) mem + 0x1010;

	*((unsigned int *) &stuck[regs->sp]) = *ip;
	regs->sp += 4;

	call->addr = *((unsigned int *) &mem[*ip + 3]);
	call->bincode = mem[*ip];
	call->mrflug0 = mem[*ip + 1];
	call->mrflug1 = mem[*ip + 2];

	if(call->mrflug0 == MOV_REG && call->mrflug1 == MOV_REG) {
		addr = &regs->regs[call->addr - REGS_BASE];
	} else if(call->mrflug0 == MOV_MEM && call->mrflug1 == MOV_REG) {
		if(call->addr - REGS_BASE <= 257) addr = (unsigned int *) &mem[regs->regs[call->addr - REGS_BASE]];
		else addr = (unsigned int *) &mem[call->addr];
	} else if(call->mrflug0 == MOV_ETC && call->mrflug1 == MOV_ETC) {
		addr = (unsigned int *) mem;
		*addr = call->addr;
	} else if(call->mrflug0 == MOV_MEM && call->mrflug1 == MOV_ETC) {
		addr = (unsigned int *) &mem[call->addr];
	}

	*ip = call->addr;
}

void run_ret(struct REGS *regs, unsigned int *ip, unsigned char *mem)
{
	regs->sp -= 4;
	*ip = *((unsigned int *) &stuck[regs->sp]);
}

void run_jmp(struct REGS *regs, unsigned int *ip, unsigned char *mem)
{
	struct IP_JMP *jmp = (struct IP_JMP *) mem + 0x1000;
	unsigned int *addr = (unsigned int *) mem + 0x1010;

	jmp->addr = *((unsigned int *) &mem[*ip + 3]);
	jmp->bincode = mem[*ip];
	jmp->mrflug0 = mem[*ip + 1];
	jmp->mrflug1 = mem[*ip + 2];

	if(jmp->mrflug0 == MOV_REG && jmp->mrflug1 == MOV_REG) {
		addr = &regs->regs[jmp->addr - REGS_BASE];
	} else if(jmp->mrflug0 == MOV_MEM && jmp->mrflug1 == MOV_REG) {
		if(jmp->addr - REGS_BASE <= 257) addr = (unsigned int *) &mem[regs->regs[jmp->addr - REGS_BASE]];
		else addr = (unsigned int *) &mem[jmp->addr];
	} else if(jmp->mrflug0 == MOV_ETC && jmp->mrflug1 == MOV_ETC) {
		addr = (unsigned int *) mem;
		*addr = jmp->addr;
	} else if(jmp->mrflug0 == MOV_MEM && jmp->mrflug1 == MOV_ETC) {
		addr = (unsigned int *) &mem[jmp->addr];
	}

	*ip = jmp->addr;
}

void run_add(struct REGS *regs, unsigned int ip, unsigned char *mem)
{
	struct IP_ADD *add = (struct IP_ADD *) mem + 0x1000;
	unsigned int *add0 = (unsigned int *) mem + 0x1010;
	unsigned int *add1 = (unsigned int *) mem + 0x1020;

	add->data1 = *((unsigned int *) &mem[ip + 9]);
	add->bincode = mem[ip];
	add->mrflug00 = mem[ip + 1];
	add->mrflug01 = mem[ip + 2];
	add->mrflug10 = mem[ip + 7];
	add->mrflug11 = mem[ip + 8];
	add->data0 = *((unsigned int *) &mem[ip + 3]);

	if(add->mrflug00 == MOV_REG && add->mrflug01 == MOV_REG) {
		add0 = &regs->regs[add->data0 - REGS_BASE];
	} else if(add->mrflug00 == MOV_MEM && add->mrflug01 == MOV_REG) {
		if(add->data0 - REGS_BASE <= 257) add0 = (unsigned int *) &mem[regs->regs[add->data0 - REGS_BASE]];
		else add0 = (unsigned int *) &mem[add->data0];
	} else if(add->mrflug00 == MOV_ETC && add->mrflug01 == MOV_ETC) {
		add0 = (unsigned int *) mem;
		*add0 = add->data0;
	} else if(add->mrflug00 == MOV_MEM && add->mrflug01 == MOV_ETC) {
		add0 = (unsigned int *) &mem[add->data0];
	}

	if(add->mrflug10 == MOV_REG && add->mrflug11 == MOV_REG) {
		add1 = &regs->regs[add->data1 - REGS_BASE];
	} else if(add->mrflug10 == MOV_MEM && add->mrflug11 == MOV_REG) {
		if(add->data1 - REGS_BASE <= 257) add1 = (unsigned int *) &mem[regs->regs[add->data1 - REGS_BASE]];
		else add1 = (unsigned int *) &mem[add->data1];
	} else if(add->mrflug10 == MOV_ETC && add->mrflug11 == MOV_ETC) {
		add1 = (unsigned int *) mem;
		*add1 = add->data1;
	} else if(add->mrflug10 == MOV_MEM && add->mrflug11 == MOV_ETC) {
		add1 = (unsigned int *) &mem[add->data1];
	}

	*add0 += *add1;
}
